# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
import inspect
import itertools
from os.path import abspath, basename, isdir, isfile, splitext

from metagenomix._io_utils import get_out_dir
from metagenomix.tools.preprocess import (
    edit, count, fastqc, fastp, cutadapt, atropos, kneaddata, filtering)
from metagenomix.tools.simka import simka
from metagenomix.tools.alignment import bowtie2, flash
from metagenomix.tools.profiling import shogun, woltka, kraken2, midas
from metagenomix.tools.phlans import metaphlan, humann, strainphlan
from metagenomix.tools.assembly import (
    pooling, spades, quast, plass, viralverify)
from metagenomix.tools.annotation import (
    prodigal, integron_finder, macsyfinder, ioncom, search, antismash, prokka)
from metagenomix.tools.binning import metawrap
from metagenomix.tools.drep import get_drep_bins, get_drep_inputs
from metagenomix.tools.metamarker import metamarker


class Commands(object):

    def __init__(self, config, databases, workflow):
        self.config = config
        self.databases = databases
        self.softs = workflow.softs
        self.outputs = {}
        self.cmds = {}
        self.args = {}
        self.pools = {}
        self.pool = None
        self.pool_bool = False
        self.inputs = None
        self.method = None
        self.soft = None
        self.sam = None
        self.dir = ''
        self.out = []
        self.io = set
        self.struc = list
        self.holistics = [
            'antismash'
            'multiqc',
            'simka',
            'quast'
            'qiita_wol',
            'mag_data',
            'metamarker',
            'woltka',
            'drep',
            'strainphlan'
        ]

    def run(self):
        for sdx, softs in enumerate(self.config.pipeline):
            self.soft = self.softs[softs[-1]]
            print('[Collecting commands] #%s: %s' % (sdx, self.soft.name))
            self.get_inputs()
            self.get_dir()
            self.generic_command()
            self.update_dirs()

    def get_inputs(self):
        """Update the `inputs` attribute of the software object."""
        if not self.soft.prev or self.soft.name == 'map__drep':
            if self.soft.params['scratch']:
                self.inputs = self.config.fastq_scratch
            else:
                self.inputs = self.config.fastq
        else:
            self.inputs = self.softs[self.soft.prev].outputs

    def get_dir(self):
        self.dir = abspath('%s/%s/after_%s' % (
            self.config.dir, self.soft.name, self.soft.prev))
        self.soft.dirs.add(self.dir)
        if self.soft.params['scratch']:
            self.dir = '${SCRATCH_FOLDER}%s' % self.dir

    def is_pool(self):
        self.pool_bool = False
        self.struc = list
        self.io = set
        if set(self.inputs) == set(self.pools):
            self.pool_bool = True
            self.struc = dict
            self.io = dict

    def generic_command(self):
        self.sam = ''
        self.is_pool()
        self.soft.io = {}
        if self.soft.name in self.holistics:
            self.prep_job()
        elif self.soft.name == 'pooling':
            self.pooling()
        else:
            for sam_or_pool in sorted(self.inputs):
                self.sam = sam_or_pool
                self.pool = sam_or_pool
                self.prep_job()
        self.register_command()

    def fill_soft_io(self):
        for i, j in itertools.product(*[['I', 'O'], ['d', 'f']]):
            if self.io == set:
                if self.sam not in self.soft.io:
                    self.soft.io[self.sam] = {}
                self.soft.io[self.sam][(i, j)] = self.outputs[
                    'io'].get((i, j), self.io())
            else:
                for k, v in self.outputs['io'].get((i, j), self.io()).items():
                    if (self.sam, k) not in self.soft.io:
                        self.soft.io[(self.sam, k)] = {}
                    self.soft.io[(self.sam, k)][(i, j)] = v

    def update_dirs(self):
        self.soft.dirs = set([
            # x.replace('${SCRATCH_FOLDER}%s' % self.dir, self.dir)
            x.replace('${SCRATCH_FOLDER}/', '/')
            for x in self.soft.dirs])

    def init_outputs(self):
        self.outputs = {'cmds': self.struc(), 'outs': self.struc(), 'dirs': [],
                        'io': {('I', 'd'): self.io(), ('I', 'f'): self.io(),
                               ('O', 'd'): self.io(), ('O', 'f'): self.io()}}

    def call_method(self):
        """Call the command-preparing method from this class (for the tools that
        are easy to deal with), or from auxillary modules located in the `tools`
        submodules path."""
        func = self.soft.name.split('_')[0]
        if func in globals():
            globals()[func](self)
        # ---------------------- TEMPORARY -----------------------
        # this is for the functions that might be defined here
        elif func in dir(self) and callable(getattr(self, func)):
            getattr(self, func)()
        else:
            raise ValueError('No method for software %s' % self.soft.name)

    def extract_data(self):
        if self.outputs.get('cmds', []):
            self.cmds[self.sam] = self.outputs['cmds']
            self.fill_soft_io()
        if self.soft.name in self.holistics:
            self.soft.outputs[self.soft.name] = self.outputs['outs']
        else:
            self.soft.outputs[self.pool] = self.outputs['outs']

    def prep_job(self):
        self.init_outputs()     # initialize data structure collection
        self.call_method()      # collect commands, outputs, io, dirs
        self.extract_data()     # fill the useful self.soft attributes

    def pooling(self):
        for pool in self.config.pooling_groups:
            self.pools[pool] = {}
            self.soft.outputs[pool] = {}
            self.soft.io[pool] = {('I', 'f'): dict(), ('I', 'd'): dict(),
                                  ('O', 'f'): dict(), ('O', 'd'): dict()}
            pooling(self, pool)

    # IN PROGRESS
    def mapping(self):
        refs = {}
        fastqs, group_fps = self.config.fastq, self.inputs[self.pool]
        if self.soft.prev == 'drep':
            for algo, fp in self.inputs[self.pool].items():
                ref_fasta = '%s.fa' % fp
                if not isfile(ref_fasta):
                    cmd = 'if [ ! -f %s ]; then cat %s/*.fa > %s; fi' % (
                        ref_fasta, fp, ref_fasta)
                    self.outputs['cmds'].setdefault(
                        (self.pool, algo), []).append(cmd)
                refs[algo] = ref_fasta
        elif self.soft.prev == 'metawrap_reassemble':
            # needs revision...
            refs = {'': glob.glob(self.inputs[self.pool][self.sam][1])}
        elif self.soft.prev == 'spades':
            refs = {group: fps[1] for group, fps in group_fps.items()}

        self.outputs['outs'] = {}
        for group, fasta in refs.items():
            self.outputs['outs'][group] = {}
            out_dir = '%s/%s' % (self.dir, self.pool)
            if group:
                out_dir += '/%s' % group
            self.outputs['dirs'].append(out_dir)
            for sam in self.pools[self.pool][group]:
                bam_out = '%s/%s.bam' % (out_dir, sam)
                self.outputs['outs'][group][sam] = bam_out
                cmd = 'minimap2'
                cmd += ' -a -x sr'
                cmd += ' %s' % fasta
                cmd += ' %s' % fastqs[sam][0]
                cmd += ' %s' % fastqs[sam][1]
                cmd += ' | samtools view -F 0x104 -b -'
                cmd += ' | samtools sort -o %s - ' % bam_out
                if not isfile(bam_out) or self.config.force:
                    self.outputs['cmds'].setdefault(
                        (self.pool, group), []).append(cmd)
                bam_out_bai = '%s.bai' % bam_out
                cmd = 'samtools index %s' % bam_out
                if not isfile(bam_out_bai) or self.config.force:
                    self.outputs['cmds'].setdefault(
                        (self.pool, group), []).append(cmd)
                io_update(self, i_f=([fasta] + fastqs[sam]),
                          o_f=[bam_out, bam_out_bai])
        print(sdfkvjjb)

    def prep_map__spades_prodigal(self):
        if 'prodigal' not in self.softs or 'mapping' not in self.softs:
            return None
        if self.softs['prodigal'].prev != 'spades':
            return None
        if self.softs['mapping'].prev != 'spades':
            return None
        prodigals_fps = self.softs['prodigal'].outputs
        sams_fps = self.softs['mapping'].outputs
        group_fps = self.inputs[self.pool]
        self.outputs['outs'] = {}
        for group, fps in group_fps.items():
            self.outputs['outs'][group] = {}
            for sam in self.pools[self.pool][group]:
                bam = sams_fps[self.pool][group][sam]
                prot = prodigals_fps[self.pool][group][1]
                out_dir = '%s/%s/%s' % (self.dir, self.pool, sam)
                out = '%s/reads.txt' % out_dir
                if not isfile(out):
                    cmd = 'pysam_reads_to_prodigal.py \\\n'
                    cmd += '-prodigal %s \\\n' % prot
                    cmd += '-bam %s \\\n' % bam
                    cmd += '-out %s\n' % out
                    self.outputs['cmds'].setdefault(
                        (self.pool, group), []).append(cmd)
                self.outputs['outs'][group][sam] = out
                io_update(self, i_f=[prot, bam, out])

    def prep_drep(self):
        self.outputs['outs'] = {}
        genomes = get_drep_bins(self.soft.prev, self.pools, self.inputs)
        for pool, group_paths in genomes.items():
            self.outputs['outs'][pool] = {}
            for stringency, sam_paths in group_paths.items():
                drep_dir = '%s/%s' % (self.dir, pool)
                if stringency:
                    drep_dir += '/%s' % stringency
                drep_in, paths = get_drep_inputs(drep_dir, sam_paths)
                io_update(self, i_f=([drep_in] + list(paths)))
                for algorithm in ['fastANI', 'ANIn']:
                    drep_out = '%s/%s' % (drep_dir, algorithm)
                    self.outputs['dirs'].append(drep_out)
                    log = '%s/log' % drep_out
                    figure = '%s/figure' % drep_out
                    data_table = '%s/data_tables' % drep_out
                    dereplicated_genomes = '%s/dereplicated_genomes' % drep_out
                    outputs = [data_table, log, figure, dereplicated_genomes]
                    self.outputs['outs'][pool][algorithm] = outputs
                    if len(glob.glob('%s/*.fa' % dereplicated_genomes)):
                        continue
                    if isdir(drep_out):
                        io_update(self, i_d=drep_out)
                    cmd = 'dRep dereplicate'
                    cmd += ' %s' % drep_out
                    cmd += ' --S_algorithm %s' % algorithm
                    cmd += ' --ignoreGenomeQuality'
                    if len(list(paths)) > 5000 and algorithm == 'fastANI':
                        cmd += ' --multiround_primary_clustering'
                        cmd += ' --greedy_secondary_clustering'
                        cmd += ' --run_tertiary_clustering'
                    cmd += ' -pa 0.9 -sa 0.98 -nc 0.3 -cm larger'
                    cmd += ' -p %s' % self.soft.params['cpus']
                    cmd += ' -g %s' % drep_in
                    self.outputs['cmds'].setdefault(pool, []).append(cmd)
                    io_update(self, o_d=outputs)

    def prep_mocat(self):
        self.cmds = {}

    def prep_mapDamage(self):
        self.cmds = {}

    def prep_metaclade(self):
        self.cmds = {}

    def prep_ezTree(self):
        self.cmds = {}

    def prep_instrain(self):
        self.cmds = {}

    def prep_map__cazy_spades(self):
        self.cmds = {}

    def prep_map__cazy_macsyfinder(self):
        self.cmds = {}

    def prep_map__spades_bins(self):
        self.cmds = {}

    def prep_map__drep(self):
        self.cmds = {}

    def prep_mag_data(self):
        self.cmds = {}

    def prep_yamb(self):
        self.cmds = {}

    def prep_summarize_yamb(self):
        self.cmds = {}

    def prep_mycc(self):
        self.cmds = {}

    def prep_gtdbtk(self):
        self.cmds = {}

    def prep_graspx(self):
        self.cmds = {}

    def prep_checkm(self):
        self.cmds = {}

    def prep_deepvirfinder(self):
        self.cmds = {}

    def prep_deepvirfinder_extractVirs(self):
        self.cmds = {}

    def prep_metaxa2(self):
        self.cmds = {}

    def prep_closed_ref_qiime1(self):
        self.cmds = {}

    def prep_phyloflash(self):
        self.cmds = {}

    def prep_wish(self):
        self.cmds = {}

    def prep_thmm(self):
        self.cmds = {}

    def prep_cazy(self):
        self.cmds = {}

    def prep_itasser(self):
        self.cmds = {}

    def register_command(self):
        self.softs[self.soft.name].cmds = dict(self.cmds)
        self.cmds = {}

    # def add_transfer_localscratch(self):
    #     pass
    #     # if self.scratch:
    #     #     self.add_transfer_localscratch()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.metagenomix import metagenomix
from metagenomix import __version__


@click.command()
@click.option(
    "-i", "--i-fastq-dir", required=True, multiple=True,
    help="Path to fastq files folder.")
@click.option(
    "-m", "--m-metadata", required=True, help="Path to the metadata file.")
@click.option(
    "-o", "--o-out-dir", required=True, help="Path to pipeline output folder.")
@click.option(
    "-n", "--p-project-name", required=True, help="Name for your project.")
@click.option(
    "-d", "--p-databases", show_default=True, help="Databases (yaml file).")
@click.option(
    "-p", "--p-pipeline", show_default=True, help="Tools to pipeline.")
@click.option(
    "-u", "--p-run-params", show_default=True, help="Server run parameters.")
@click.option(
    "-c", "--p-co-assembly", show_default=True,
    help="Metadata column(s) defining the co-assembly groups (yaml file).")
@click.option(
    "-s", "--p-strains", show_default=True, help="Species for strain-level"
                                                 " analyses (yaml file).")
@click.option(
    "-M", "--p-modules", show_default=True, help="modules to use per software"
                                                 " analyses (yaml file).")
@click.option(
    "-a", "--p-account", show_default=False, default=None,
    help="User account on HPC")
@click.option(
    "--force/--no-force", default=False, show_default=True,
    help="Force the re-writing of scripts for all commands"
         "(default is to not re-run if output file exists).")
@click.option(
    "--jobs/--no-jobs", default=True, show_default=True,
    help="Whether to prepare Torque jobs from scripts.")
@click.option(
    "--torque/--no-torque", default=False, show_default=True,
    help="Whether to prepare Torque jobs instead of Slurm.")
@click.option(
    "-l", "--localscratch", type=int, show_default=False, default=None,
    help="Use localscratch with the provided memory amount (in GB)")
@click.option(
    "--scratch/--no-scratch", default=False, show_default=True,
    help="Use the scratch folder to move files and compute")
@click.option(
    "--userscratch/--no-userscratch", default=False, show_default=True,
    help="Use the userscratch folder to move files and compute")
@click.option(
    "--show-params/--no-show-params", default=False, show_default=False,
    help="Show all possible parameters for all tools of your pipeline.")
@click.option(
    "--show-pfams/--no-show-pfams", default=False, show_default=False,
    help="Show terms for which Pfam HMM models were already extracted before.")
@click.option(
    "--purge-pfams", "--no-purge-pfams", default=None, show_default=False,
    help="Remove terms for Pfam HMM models that were already extracted before.")
@click.option(
    "--verbose/--no-verbose", default=False, show_default=True,
    help="Whether to show expected input and outputs and other tools' details.")
@click.option(
    "--dev/--no-dev", default=False, show_default=True,
    help="For development...")
@click.version_option(__version__, prog_name="metagenomix")


def standalone_metagenomix(
        m_metadata,
        i_fastq_dir,
        o_out_dir,
        p_project_name,
        p_co_assembly,
        p_databases,
        p_pipeline,
        p_run_params,
        p_strains,
        p_modules,
        p_account,
        force,
        jobs,
        torque,
        localscratch,
        scratch,
        userscratch,
        show_params,
        show_pfams,
        purge_pfams,
        verbose,
        dev
):

    metagenomix(
        meta_fp=m_metadata,
        fastq_dirs=i_fastq_dir,
        output_dir=o_out_dir,
        project=p_project_name,
        coassembly_yml=p_co_assembly,
        databases_yml=p_databases,
        pipeline_tsv=p_pipeline,
        user_params_yml=p_run_params,
        strains_yml=p_strains,
        modules_yml=p_modules,
        force=force,
        jobs=jobs,
        torque=torque,
        account=p_account,
        localscratch=localscratch,
        scratch=scratch,
        userscratch=userscratch,
        show_params=show_params,
        show_pfams=show_pfams,
        purge_pfams=purge_pfams,
        verbose=verbose,
        dev=dev
    )


if __name__ == "__main__":
    standalone_metagenomix()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import argparse
import pandas as pd


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', nargs=1, required=True, help='Upper triangular')
    parse = parser.parse_args()
    arguments = vars(parse)
    return arguments


def get_samples(mat_fp: str) -> list:
    """Get the sample names which are in the
    first line of the upper triangular matrix.

    Parameters
    ----------
    mat_fp : str
        Input path to the upper triangular matrix.

    Returns
    -------
    first_line : list
        Sample names.
    """
    with open(mat_fp) as f:
        for ldx, line in enumerate(f):
            first_line = line.strip().split(';')[1:]
            return first_line


def get_matrix_in_pandas(mat_fp: str) -> pd.DataFrame:
    """Read the upper triangular matrix into pandas.

    Parameters
    ----------
    mat_fp : str
        Input path to the upper triangular matrix.

    Returns
    -------
    mat_pd : pd.DataFrame
        Upper triangular matrix table.
    """
    print('reading matrix in pandas...')
    mat_pd = pd.read_csv(
        mat_fp,
        skiprows=1,
        header=None,
        index_col=0,
        sep=';',
        low_memory=False,
        dtype=str
    )
    return mat_pd


def write_matrix_from_pandas(mat_pd: pd.DataFrame, mat_fp: str, samples: list):
    """Write the matrix as symmetric table.

    Parameters
    ----------
    mat_pd : pd.DataFrame
        Upper triangular matrix table.
    mat_fp : str
        Input path to the upper triangular matrix.
    samples : list
        Sample names.
    """
    mat_o = mat_fp.replace('.csv', '_sym.tsv')
    with open(mat_o, 'w') as o:
        o.write('\t%s\n' % '\t'.join(samples))
        for idx, row in enumerate(mat_pd.values):
            for jdx in range(len(row)):
                if jdx < idx:
                    row[jdx] = mat_pd.iloc[jdx, idx]
            o.write('%s\t%s\n' % (samples[idx], '\t'.join(row)))


if __name__ == '__main__':
    args = get_args()
    fp = args['i'][0]
    samples = get_samples(fp)
    mat_pd = get_matrix_in_pandas(fp)
    write_matrix_from_pandas(mat_pd, fp, samples)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import unittest

import pandas as pd
import pkg_resources
from pandas.testing import assert_frame_equal, assert_series_equal

from metagenomix._metadata import read_metadata, get_cur_type, get_first_column

FOLDER = pkg_resources.resource_filename('metagenomix', 'tests/unittests')


class ReadMetadata(unittest.TestCase):

    def setUp(self) -> None:

        self.empty_meta_fp = '%s/metadata/empty_meta.txt' % FOLDER
        with open(self.empty_meta_fp, 'w'):
            pass

        self.only_header_fp = '%s/metadata/only_header.txt' % FOLDER
        with open(self.only_header_fp, 'w') as o:
            o.write('sample_name\tdepth\tlight')
        self.only_header = pd.DataFrame(
            {'sample_name': [], 'depth': [], 'light': []})

        self.meta_fp = '%s/metadata/meta.txt' % FOLDER
        with open(self.meta_fp, 'w') as o:
            o.write('sample_name\tdepth\tlight\n')
            o.write('A\t10\tyes\nB\t2000\tno\nX\t0\tyes')
        self.meta = pd.DataFrame({'sample_name': ['A', 'B', 'X'],
                                  'depth': ['10', '2000', '0'],
                                  'light': ['yes', 'no', 'yes']})

        self.alt_header_fp = '%s/metadata/alt_header.txt' % FOLDER
        with open(self.alt_header_fp, 'w') as o:
            o.write('some_name\tdepth\tlight\n')
            o.write('A\t10\tyes\nB\t2000\tno\nX\t0\tyes')
        self.alt_header = self.meta.copy().rename(
            columns={'sample_name': 'some_name'})

    def test_read_metadata(self):
        with self.assertRaises(IOError) as e:
            read_metadata(self.empty_meta_fp)
        self.assertEqual(
            'File "%s" is empty' % self.empty_meta_fp, str(e.exception))

        meta = read_metadata(self.only_header_fp)
        self.assertEqual(str(meta), str(self.only_header))
        self.assertEqual(meta.values.tolist(), self.only_header.values.tolist())

        meta = read_metadata(self.meta_fp)
        assert_frame_equal(meta, self.meta)

        meta = read_metadata(self.alt_header_fp)
        assert_frame_equal(meta, self.meta)

    def test_get_first_column(self):
        first_column = get_first_column(self.meta_fp)
        self.assertEqual(first_column, 'sample_name')
        first_column = get_first_column(self.alt_header_fp)
        self.assertEqual(first_column, 'some_name')
        first_column = get_first_column(self.only_header_fp)
        self.assertEqual(first_column, 'sample_name')
        with self.assertRaises(IOError) as e:
            get_first_column(self.empty_meta_fp)
        self.assertEqual(
            'File "%s" is empty' % self.empty_meta_fp, str(e.exception))

    def tearDown(self) -> None:
        os.remove(self.meta_fp)
        os.remove(self.only_header_fp)
        os.remove(self.alt_header_fp)
        os.remove(self.empty_meta_fp)


class CheckType(unittest.TestCase):

    def setUp(self) -> None:

        self.short_fp = '%s/short.fa' % FOLDER
        self.nucl_fp = '%s/nucl.fa' % FOLDER
        self.prot_fp = '%s/prot.fa' % FOLDER
        self.neither_fp = '%s/neither.fa' % FOLDER
        with open(self.short_fp, 'w') as o:
            o.write('>header\n')
            o.write('ACTGTGATTG\n')

        with open(self.nucl_fp, 'w') as o:
            o.write('>header\n')
            o.write('ACTGTGATTGCANTAN\n')

        with open(self.prot_fp, 'w') as o:
            o.write('>header\n')
            o.write('POSYYRTWBNHCCRCG\n')

        with open(self.neither_fp, 'w') as o:
            o.write('Nothing\n')
            o.write('Special\n')

    def test_get_cur_type(self):
        ret = get_cur_type(self.short_fp)
        self.assertEqual(ret, 'nucl')
        ret = get_cur_type(self.nucl_fp)
        self.assertEqual(ret, 'nucl')
        ret = get_cur_type(self.prot_fp)
        self.assertEqual(ret, 'prot')
        ret = get_cur_type(self.neither_fp)
        self.assertEqual(ret, 'prot')

    def tearDown(self) -> None:
        os.remove(self.short_fp)
        os.remove(self.nucl_fp)
        os.remove(self.prot_fp)
        os.remove(self.neither_fp)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
from os.path import dirname, isfile
from metagenomix._io_utils import io_update


def metaphlan(self) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    tmpdir = '$TMPDIR/metaphlan_%s' % self.sam
    tmp_cmd = ['mkdir -p %s\n' % tmpdir]

    bowtie2out = '%s/bowtie2/%s.bowtie2.bz2' % (self.dir, self.sam)

    outs = profiling(self, bowtie2out, tmpdir)
    self.outputs['outs'] = [bowtie2out] + list(outs)
    self.outputs['dir'] = [dirname(x) for x in self.outputs['outs']]

    reads = get_read_count(self)
    if reads:
        analyses(self, bowtie2out, reads)
    if self.outputs['cmds']:
        self.outputs['cmds'] = tmp_cmd + self.outputs['cmds']


def get_read_count(self) -> str:
    """Get the number of reads in the current sample,
    or an empty string if the counting did not happen yet.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name

    Returns
    -------
    reads : str
        Number of reads.
    """
    reads = ''
    reads_fps = self.softs['count'].outputs
    if self.sam in reads_fps and isfile(reads_fps[self.sam][0]):
        with open(reads_fps[self.sam][0]) as f:
            for line in f:
                reads = line[-1].strip().split(',')[-1]
                break
    return reads


def analyses(self, bowtie2out: str, reads: str) -> None:
    """Get the taxonomic analyses Metaphlan command.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .config
            Configurations
    bowtie2out : str
        Input type was a bowtie2 output of the same tool
    reads : str
        Number of reads in the current sample
    """
    tax_levs = ['a', 'k', 'p', 'c', 'o', 'f', 'g', 's']
    for analysis in ['rel_ab',
                     'rel_ab_w_read_stats',
                     'clade_profiles',
                     'marker_ab_table',
                     'marker_pres_table',
                     'clade_specific_strain_tracker']:
        out = '%s/%s' % (self.dir, analysis)
        for tax_lev in tax_levs:
            rad = '%s/%s_t-%s' % (out, self.sam, tax_lev)
            cmd = 'metaphlan'
            cmd += ' %s' % bowtie2out
            cmd += ' -t %s' % analysis
            cmd += ' --tax_lev %s' % tax_lev
            cmd += ' --input_type bowtie2out'
            cmd += ' --sample_id %s' % self.sam
            cmd += ' --sample_id_key sample_name'
            cmd += ' --nproc %s' % self.soft.params['cpus']
            if reads and analysis == 'marker_ab_table':
                cmd += ' --nreads %s' % reads
            if analysis == 'clade_specific_strain_tracker':
                for strain in self.config.strains:
                    strain_cmd = cmd
                    strain_cmd += ' --clade %s' % strain.replace(' ', '_')
                    clade_out = '%s_%s.tsv' % (rad, strain.replace(' ', '_'))
                    strain_cmd += ' --output_file %s' % clade_out
                    if self.config.force or not isfile(clade_out):
                        self.outputs['cmds'].append(strain_cmd)
                        io_update(self, o_f=clade_out)
            else:
                ab_out = '%s.tsv' % rad
                cmd += ' --output_file %s' % ab_out
                if self.config.force or not isfile(ab_out):
                    self.outputs['cmds'].append(cmd)
                    io_update(self, o_f=ab_out)


def profiling(self, bowtie2out: str, tmpdir: str) -> tuple:
    """Get the taxonomic profiling metaphlan command.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    bowtie2out : str
        Input type was a bowtie2 output of the same tool
    tmpdir : str
        Path to a temporary directory

    Returns
    -------
    sam_out : str
        Alignment output
    profile_out : str
        Metaphlan taxonomic profile output
    """
    sam_out = '%s/sams/%s.sam.bz2' % (self.dir, self.sam)
    profile_out = '%s/profiles/%s.tsv' % (self.dir, self.sam)

    if isfile(profile_out):
        io_update(self, i_f=bowtie2out)
    else:
        io_update(self, o_f=profile_out)
        cmd = 'metaphlan'
        if isfile(bowtie2out):
            io_update(self, i_f=bowtie2out)
            cmd += ' %s' % bowtie2out
            cmd += ' --input_type bowtie2out'
        else:
            io_update(self, i_f=self.inputs[self.sam])
            cmd += ' %s' % ','.join(self.inputs[self.sam])
            cmd += ' --input_type fastq'
            cmd += ' --samout %s' % sam_out
            cmd += ' --bowtie2out %s' % bowtie2out
            cmd += ' --bowtie2db %s' % self.databases.paths['metaphlan']
        cmd += ' --tmp_dir %s' % tmpdir
        cmd += ' --output_file %s' % profile_out
        cmd += ' --nproc %s' % self.soft.params['cpus']
        cmd += ' --sample_id_key %s' % self.sam
        self.outputs['cmds'].append(cmd)

    return sam_out, profile_out


def get_profile(self) -> list:
    """Get the taxonomic profiles if there are some, for humann to only run
    of the taxa from these profiles.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases

    Returns
    -------
    profiles_dbs_outs : list
        Element are lists:
            profiles : list
                Paths to the taxonomic profile of the current sample.
            dbs : list
                Paths to the metaphlan database.
            outs : list
                Paths to the output folder.
    """
    profiles, dbs, outs = [], [], []
    db_path = self.databases.paths['humann']
    params = self.soft.params
    if params['profiles']:
        for profile_name, profile_fp in params['profiles'].items():
            # `profile_fp` must be "path/to/<sample>/profile_file.txt"
            profile_sams = {x.split('/')[-2]: x for x in glob.glob(profile_fp)}
            if self.sam in profile_sams:
                profiles.append(profile_sams[self.sam])
                io_update(self, i_f=profile_sams[self.sam])
                dbs.append('')
                outs.append(
                    self.dir + '/profile_%s_%s_id%s_e%s_qC%s_sC%s_bypss%s' % (
                        profile_name,
                        params['uniref'],
                        params['id_thresh'],
                        params['evalue'],
                        params['query_coverage'],
                        params['subject_coverage'],
                        params['skip_translated']))
    else:
        profiles.append('')
        dbs.append(db_path)
        outs.append(self.dir + '/no_profile')
    profiles_dbs_outs = [profiles, dbs, outs]
    return profiles_dbs_outs


def get_cmd(self, profile: str, db: str, out: str, base: str, ali: str):
    """Get the humann command line.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .soft.params
            Parameters for humann
    profile : str
        Path to the taxonomic profile of the current sample.
    db : str
        Path to the metaphlan database.
    out : str
        Path to the output folder.
    base : str
        Path to fastq file.
    ali : str
        Path to temporary, alignment sam file

    Returns
    -------
    cmd : str
        humann command line.
    """
    cmd = 'cat %s > %s\n' % (' '.join(self.inputs[self.sam]), base)
    cmd += 'humann'
    cmd += ' --input %s' % base
    cmd += ' --output %s' % out
    cmd += ' --prescreen-threshold %s' % self.soft.params['prescreen_threshold']
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --search-mode %s' % self.soft.params['uniref']
    if len(db):
        cmd += ' --metaphlan-options '
        cmd += '" --mpa_pkl %s/mpa_v20_m200.pkl' % db
    cmd += ' --bowtie2db %s"' % db
    cmd += ' --identity-threshold %s' % self.soft.params['identity']
    cmd += ' --evalue %s' % self.soft.params['evalue']
    cmd += ' --translated-query-coverage-threshold'
    cmd += ' %s' % self.soft.params['subject_cov']
    cmd += ' --translated-query-coverage-threshold'
    cmd += ' %s' % self.soft.params['subject_cov']
    if profile:
        cmd += ' --taxonomic-profile %s' % profile
    if self.soft.params['skip_translated']:
        cmd += ' --bypass-translated-search'
    if isfile(ali):
        cmd += ' -r'
    cmd += ' --nucleotide-database %s' % self.soft.params['nucleotide_db']
    cmd += ' --protein-database  %s\n' % self.soft.params['protein_db']
    return cmd


def renorm(input_fp: str, relab_fp: str) -> str:
    """Normalize the reads counts into proportions.

    Parameters
    ----------
    input_fp : str
        Path to the input file.
    relab_fp : str
        Path to the output file.

    Returns
    -------
    cmd : str
        Normalization command line.
    """
    cmd = 'humann2_renorm_table'
    cmd += ' --input %s' % input_fp
    cmd += ' --output %s' % relab_fp
    cmd += ' -u relab\n'
    return cmd


def get_ali(self, out: str):
    """

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .outputs : dict
            All outputs
    out : str


    Returns
    -------
    ali : str
        Path to the alignment .sam file
    """
    ali = '%s/%s_tmp/%s_bowtie2_aligned.sam' % (out, self.sam, self.sam)
    if not self.config.force and isfile(ali):
        io_update(self, i_f=ali)
    else:
        io_update(self, o_f=ali, o_d=out)
    return ali


def get_outputs(self, profile: str, db: str, out: str, ali: str) -> tuple:
    """Get the profiling output files at genus, pathway and coverage levels.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .config
            Configurations
    profile : str
        Path to the taxonomic profile of the current sample.
    db : str
        Path to the metaphlan database.
    out : str
        Path to the output folder.
    ali : str
        Path to temporary, alignment sam file

    Returns
    -------
    gen : str
        Path to the genus abundance profile output file
    pwy : str
        Path to the pathway abundance profile output file
    cov : str
        Path to the pathway coverage profile output file
    """
    base = '%s/%s.fastq.gz' % (out, self.sam)
    gen = '%s/%s_genefamilies.tsv' % (out, self.sam)
    pwy = '%s/%s_pathabundance.tsv' % (out, self.sam)
    cov = '%s/%s_pathcoverage.tsv' % (out, self.sam)
    if not self.config.force and isfile(gen) and isfile(pwy) and isfile(cov):
        io_update(self, i_f=[gen, pwy])
    else:
        self.outputs['cmds'].append(get_cmd(self, profile, db, out, base, ali))
        io_update(self, o_f=[gen, pwy, cov])
    return gen, pwy, cov


def humann(self) -> None:
    """Create command lines for humann for the current database's species focus.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    for profile, db, out in zip(*get_profile(self)):
        ali = get_ali(self, out)
        gen, pwy, cov = get_outputs(self, ali, profile, db, out)
        gen_relab = '%s/%s_genefamilies_relab.tsv' % (out, self.sam)
        pwy_relab = '%s/%s_pathabundance_relab.tsv' % (out, self.sam)
        if self.config.force or not isfile(gen_relab) or not isfile(pwy_relab):
            self.outputs['cmds'].append(renorm(gen, gen_relab))
            self.outputs['cmds'].append(renorm(pwy, pwy_relab))
            io_update(self, o_f=[gen_relab, pwy_relab])
        self.outputs['dirs'].append(out)
        self.outputs['outs'].append([out, ali, gen, gen_relab,
                                     pwy, pwy_relab, cov])


def get_sample_to_marker(self, sam_dir, markers_dir):
    if self.config.force or len(glob.glob('%s/*.sam.bz2' % sam_dir)):
        cmd = 'sample2markers.py'
        cmd += ' -i %s/*.sam.bz2' % sam_dir
        cmd += ' -o %s' % markers_dir
        cmd += ' -n %s' % self.soft.params['cpus']
        self.outputs['cmds'].append(cmd)


def extract_markers(self, st, strain_dir, db_dir):
    if self.config.force or not isfile('%s/%s.fna' % (strain_dir, st)):
        cmd = 'extract_markers.py'
        cmd += ' -c %s' % st
        cmd += ' -o %s' % db_dir
        self.outputs['cmds'].append(cmd)


def strainphlan(self) -> None:
    """Create command lines for strainphlan.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    bt2_fp, sam_fp = self.inputs[list(self.inputs.keys())[0]][:2]
    sam_dir = dirname(sam_fp)
    tmp_dir = '$TMPDIR/strainphlan'
    db_dir = '%s/db_markers' % self.dir
    meta_dir = '%s/metadata' % self.dir
    markers_dir = '%s/consensus_markers' % self.dir
    wol_pd = self.databases.paths['wol']['taxonomy']

    self.outputs['outs'] = dict({})
    self.outputs['dirs'].extend([db_dir, meta_dir])
    self.outputs['cmds'].append('mkdir -p %s' % tmp_dir)
    io_update(self, i_d=sam_dir, o_d=[markers_dir, db_dir, meta_dir])

    get_sample_to_marker(self, sam_dir, markers_dir)

    for strain_group, strains in self.config.strains.items():
        strain_dir = '%s/%s' % (db_dir, strain_group)
        self.outputs['dirs'].append(strain_dir)
        for strain in [x.replace(' ', '_') for x in strains]:
            if strain[0] != 's':
                continue
            extract_markers(self, strain, strain_dir, db_dir)
            odir = '%s/output/%s' % (self.dir, strain_group)
            self.outputs['dirs'].append(odir)
            self.outputs['outs'][(strain_group, strain)] = odir
            io_update(self, o_d=odir)
            tree = '%s/RAxML_bestTree.%s.StrainPhlAn3.tre' % (odir, strain)
            if self.config.force or not isfile(tree):
                cmd = 'strainphlan'
                cmd += ' -s %s/*.pkl' % markers_dir
                cmd += ' -m %s/%s.fna' % (db_dir, strain)
                if strain in wol_pd['species']:
                    fna = '%s/%s.fna.bz2' % (
                        self.databases.paths['wol']['fna'],
                        wol_pd.loc[wol_pd['species'] == strain, 0])
                    io_update(self, i_f=fna)
                    cmd += ' -r %s' % fna
                cmd += ' -o %s' % odir
                cmd += ' -n %s' % self.soft.params['cpus']
                cmd += ' -c %s' % strain
                cmd += ' --mutation_rates'
                self.outputs['cmds'].append(cmd)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd


def filter_table(preval: str, abund: str,
                 tsv_pd: pd.DataFrame) -> pd.DataFrame:
    preval = float(preval)
    abund = float(abund)
    if preval + abund == 0:
        return tsv_pd
    tsv_filt_pd = tsv_pd.copy()
    # get the min number of samples based on prevalence percent
    if preval < 1:
        n_perc = tsv_pd.shape[1] * preval
    else:
        n_perc = preval
    # abundance filter in terms of min reads counts
    tsv_pd_perc = tsv_filt_pd.copy()
    if abund < 1:
        tsv_pd_perc = tsv_filt_pd / tsv_filt_pd.sum()
    tsv_filt_pd = tsv_filt_pd.loc[(tsv_pd_perc > abund).sum(1) > n_perc, :]
    tsv_filt_pd = tsv_filt_pd.loc[tsv_filt_pd.sum(1) > 0,
                                  tsv_filt_pd.sum(0) > 0]
    return tsv_filt_pd


def write_filtered_tsv(tsv_out: str, tsv_pd: pd.DataFrame) -> None:
    tsv_sams_col = tsv_pd.reset_index().columns[0]
    tsv_pd = tsv_pd.reset_index().rename(
        columns={tsv_sams_col: 'Feature ID'}).set_index('Feature ID')
    tsv_pd.reset_index().to_csv(tsv_out, index=False, sep='\t')


def get_unique_filterings(songbird_filtering):
    unique_filterings = {}
    for filt_name, dat_d in songbird_filtering[''].items():
        for dat, (preval, abund) in dat_d.items():
            unique_filterings.setdefault(dat, set()).add(
                (filt_name, preval, abund))
    return unique_filterings



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
import pkg_resources
from os.path import isdir, isfile
from metagenomix._io_utils import mkdr, io_update

RESOURCES = pkg_resources.resource_filename("metagenomix", "resources/scripts")


def get_simka_input(self) -> str:
    """Write the file containing the paths to each sample.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .inputs : dict
            Input files

    Returns
    -------
    sim_out : str
        Path to the file to write.
    """
    out_dir = self.dir.replace('${SCRATCH_FOLDER}', '')
    sim_out = '%s/samples_files.txt' % out_dir
    mkdr(sim_out, True)
    with open(sim_out, 'w') as o:
        for sam in self.inputs:
            fs = self.inputs[sam]
            if 'after_None' not in out_dir:
                fs = ['%sq' % x[:-1] if 'notCombined_' in x else x for x in fs]
            o.write('%s: %s\n' % (sam, '; '.join(fs)))
    return sim_out


def simka_cmd(self, smin: bool, sim_in: str, out_dir: str,
              k: int, n: int) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
        .config
            Configurations
    smin : bool
        Whether to use SimkaMin (True) or base Simka (False)
    sim_in : str
        Input file containing to the fastq file paths for Simka
    out_dir : str
        Output folder for Simka
    k : int
        Length of the k-mer
    n : int
        Number of sequences to inject in the Simka analysis

    Returns
    -------
    cmd : str
        Simka command line
    """
    cmd = ''
    if isdir('%s/simkamin' % out_dir):
        cmd = 'rm -rf %s/simkamin\n' % out_dir
    if smin:
        if not self.config.force:
            if isfile('%s/mat_abundance_braycurtis.csv' % out_dir):
                return ''
            elif isfile('%s/mat_abundance_braycurtis.csv.gz' % out_dir):
                return ''
        cmd += simka_min_cmd(self.soft.params, sim_in, out_dir, k, str(n))
    else:
        cmd += simka_base_cmd(self.soft.params, sim_in, out_dir, k, str(n))
    return cmd


def simka_min_cmd(params: dict, sim_in: str, out_dir: str,
                  k: int, n: str) -> str:
    """Write the Simka command for the SimkaMin algorithm.

    Parameters
    ----------
    params : dict
        Simka parameters
    sim_in : str
        Input file containing to the fastq file paths for Simka.
    out_dir : str
        Output folder for Simka.
    k : int
        Length of the k-mer.
    n : str
        Number of sequences to inject in the Simka analysis.

    Returns
    -------
    cmd : str
        Simka command line.
    """
    cmd = 'python %s/simkaMin/simkaMin.py' % params['path']
    cmd += ' -bin %s/build/bin/simkaMinCore' % params['path']
    cmd += ' -in %s -out %s' % (sim_in, out_dir)
    cmd += ' -kmer-size %s' % str(k)
    cmd += ' -max-reads %s' % n
    cmd += ' -nb-kmers 50000'
    if int(params['mem_num']) == 1:
        mem = '1'
    else:
        mem = str((params['mem_num'] - 1))
    if params['mem_dim'].lower()[0] == 'g':
        mem += '000'
    cmd += ' -max-memory %s' % mem
    cmd += ' -min-read-size 100'
    cmd += ' -filter'
    cmd += ' -nb-cores %s\n' % params['cpus']
    cmd += 'rm -rf %s/simkamin' % out_dir
    return cmd


def simka_base_cmd(
        params: dict, sim_in: str, out_dir: str, k: int, n: str) -> str:
    """Write the Simka command for the Simka base algorithm.

    Parameters
    ----------
    params : dict
        Simka parameters
    sim_in : str
        Input file containing to the fastq file paths for Simka.
    out_dir : str
        Output folder for Simka.
    k : int
        Length of the k-mer.
    n : str
        Number of sequences to inject in the Simka analysis.

    Returns
    -------
    cmd : str
        Simka command line.
    """
    cmd = '%s/build/bin/simka' % params['path']
    cmd += ' -in %s' % sim_in
    cmd += ' -out %s' % out_dir
    cmd += ' -out-tmp %s_tmp' % out_dir
    cmd += ' -abundance-min 5'
    cmd += ' -kmer-size %s' % int(k)
    cmd += ' -max-reads %s' % n
    cmd += ' -data-info'
    cmd += ' -simple-dist'
    cmd += ' -nb-cores %s' % params['cpus']
    cmd += ' -max-count %s' % params['cpus']
    cmd += ' -max-memory %s\n' % str((int(params['mem_num'][0])*1000)-1000)
    cmd += 'rm -rf %s_tmp' % out_dir
    return cmd


def simka_pcoa_cmd(mat: str, config) -> str:
    """Write the Simka command for the pcoa based on the distance matrices.

    Parameters
    ----------
    mat : str
        Path to the input matrix.
    config
        Configurations

    Returns
    -------
    cmd : str
        Simka command line.
    """
    cmd = ''
    if not config.force and 'sym' in mat:
        return cmd
    if mat.endswith('gz'):
        mat_fp = mat.replace('.csv.gz', '.csv')
        if not isfile(mat_fp):
            cmd += 'gunzip %s\n' % mat
    else:
        mat_fp = mat

    mat_o = mat_fp.replace('.csv', '_sym.tsv')
    if config.force or isfile(mat_o):
        cmd += 'python %s/symmetrize_simka_matrix.py' % RESOURCES
        cmd += ' -i %s\n' % mat_fp

    mat_dm = mat_o.replace('.tsv', '_dm.qza')
    if config.force or not isfile(mat_dm):
        cmd += 'qiime tools import'
        cmd += ' --input-path %s' % mat_o
        cmd += ' --output-path %s' % mat_dm
        cmd += ' --type DistanceMatrix\n'

    mat_pcoa = mat_dm.replace('.qza', '_pcoa.qza')
    if config.force or not isfile(mat_pcoa):
        cmd += 'qiime diversity pcoa'
        cmd += ' --i-distance-matrix %s' % mat_dm
        cmd += ' --o-pcoa %s\n' % mat_pcoa

    mat_pcoa_dir = mat_pcoa.replace('.qza', '')
    mat_pcoa_txt = mat_pcoa.replace('.qza', '.txt')
    if config.force or not isfile(mat_pcoa_txt):
        cmd += 'qiime tools export'
        cmd += ' --input-path %s' % mat_pcoa
        cmd += ' --output-path %s\n' % mat_pcoa_dir
        cmd += 'mv %s/ordination.txt %s\n' % (mat_pcoa_dir, mat_pcoa_txt)
        cmd += 'rm -rf %s\n' % mat_pcoa_dir

    mat_emp = mat_pcoa.replace('.qza', '_emp.qzv')
    if config.force or not isfile(mat_emp):
        cmd += 'qiime emperor plot'
        cmd += ' --i-pcoa %s' % mat_pcoa
        cmd += ' --m-metadata-file %s' % config.meta_fp
        cmd += ' --o-visualization %s' % mat_emp
    return cmd


def simka(self) -> None:
    """Create command lines for Simka.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    input_file = get_simka_input(self)
    io_update(self, i_f=input_file)
    smin = self.soft.params['simkaMin']
    for k in map(int, self.soft.params['kmer']):
        for n in map(int, self.soft.params['log_reads']):
            out_d = '%s/k%s/n%s' % (self.dir, k, n)
            cmd = simka_cmd(self, smin, input_file, out_d, k, n)
            if cmd:
                self.outputs['dirs'].append(out_d)
                self.outputs['cmds'].append(cmd)
                io_update(self, o_d=out_d)
            for mdx, mat in enumerate(glob.glob('%s/mat_*.csv*' % out_d)):
                cmd = simka_pcoa_cmd(mat, self.config)
                if cmd:
                    self.outputs['cmds'].append(cmd)
                    io_update(self, o_d=out_d)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import re
import sys
import yaml
import pandas as pd
from os.path import abspath, basename, dirname, isfile


def read_yaml(
        file_path: str
) -> dict:
    """Simply reads a yaml and return its contents in a dictionary structure.

    Parameters
    ----------
    file_path: str
        Path to a yaml file.

    Returns
    -------
    yaml_dict : dict
        Dictionary returned by reading the yaml file (could be empty).
    """
    yaml_dict = {}
    if file_path:
        if not isfile(file_path):
            raise IOError('No yaml file "%s"' % file_path)
        with open(file_path) as yaml_handle:
            try:
                yaml_dict = yaml.load(yaml_handle, Loader=yaml.FullLoader)
            except AttributeError:
                yaml_dict = yaml.load(yaml_handle)
    return yaml_dict


def fill_fastq(fastq_paths: list, sams: set):
    """Populate a `fastq` dict with for each sample (keys) the list of
    fastq file paths (values), which would be either of length 1 if there
    is only one fastq file for the sample, or of length 2 if there are
    two, paired fastq file paths.

    Notes
    -----
    In fact, the values could have a length 4 since all *fastq* files are
    considered and there could be both `.fastq` and `.fastq.gz` file in the
    input folder. Only the `.fastq.gz` files would be used in a later stage.

    Parameters
    ----------
    fastq_paths : list
        All fastq files in the input folder
    sams : set
        All unique sample names that have a metadata

    Returns
    -------
    fastq : dict
        Fastq file path(s) per sample
    """
    fastq = {}
    for fastq_path in sorted(fastq_paths):
        for sam in sams:
            if basename(fastq_path).startswith(sam):
                if sam in fastq:
                    fastq[sam].append(abspath(fastq_path))
                else:
                    fastq[sam] = [abspath(fastq_path)]
                break
    return fastq


def mkdr(path: str, is_file: bool = False) -> None:
    """Creates a folder is does not exist yet.

    Parameters
    ----------
    path : str
        Path of the file or folder to create.
    is_file : bool
        True is the path is a file path, False is it is a folder path.
    """
    if is_file:
        os.makedirs(dirname(path), exist_ok=True)
    else:
        os.makedirs(path, exist_ok=True)


def get_pfam_wget_cmd(pfam_dir: str) -> str:
    """Get the validation (even through downloading) that the HMM file is there.

    Parameters
    ----------
    pfam_dir : str
        Path to the folder that should contain the "Pfam-A.*" files

    Returns
    -------
    cmd : str
        Command line to run to download the database
    """
    cmd = ''
    for ext in ['hmm', 'fasta', 'hmm.dat']:
        fp = '%s/Pfam-A.%s' % (pfam_dir, ext)
        if not isfile(fp):
            gz = 'pub/databases/Pfam/releases/Pfam35.0/Pfam-A.%s.gz' % ext
            cmd += 'wget -O %s http://ftp.ebi.ac.uk/%s\n' % (fp, gz)
            cmd += 'gunzip %s\n' % fp
    return cmd


def get_hmm_dat(pfam_dir: str) -> pd.DataFrame:
    """Get as table the contents of
    http://ftp.ebi.ac.uk/pub/databases/Pfam/releases/Pfam35.0/Pfam-A.hmm.dat.gz

    Parameters
    ----------
    pfam_dir : str
        Path to the folder that should contain the "Pfam-A.*" files

    Returns
    -------
    pfam_pd : pd.DataFrame
        Pfam dat reformatted as table.
    """
    tsv = '%s/Pfam-A.hmm.dat.tsv' % pfam_dir
    if not isfile(tsv):
        records = []
        cur = {}
        with open('%s/Pfam-A.hmm.dat' % pfam_dir) as f:
            for line in f:
                line_decode = line.strip()
                if line_decode.startswith('//'):
                    records.append(cur)
                    cur = {}
                elif line_decode.startswith('#=GF'):
                    cur[line_decode[5:7]] = line_decode[7:].strip()
        pfam_pd = pd.DataFrame(records)
        pfam_pd.to_csv(tsv, index=False, sep='\t')
    else:
        pfam_pd = pd.read_table(tsv)
    return pfam_pd


def get_pfams_cmd(hmm: str, term_pd: pd.DataFrame,
                  term: str, odir: str) -> tuple:
    """Collect the hmm profiles and diamond database for different term targets.

    Parameters
    ----------
    hmm : str
        Pfam-A.hmm file.
    term_pd : pd.DataFrame
        Data frame containing the Pfam-A.hmm.dat file, but reformatted above.
    odir : str
        Directory where all Pfam models will be kept.
    term : str
        Term to search among the Pfam hmms.

    Returns
    -------
    pfams : dict
        [.hmm, .dmnd] files for each term.
    cmd : str
        Command to hmmfetch these term's hmms.
    """
    comp = re.compile("[\\ \-\",()%':&/.\[\]]")
    cmd = ''
    pfams = {}
    pfam_out = '%s/%s' % (odir, comp.sub('_', term))
    mkdr(pfam_out)
    for r, row in term_pd.iterrows():
        acc, desc = row['AC'], comp.sub('_', row['DE'])
        name = '%s__%s' % (acc, desc)
        hmm_fp = '%s/%s.hmm' % (pfam_out, name)
        if not isfile(hmm_fp):
            cmd += 'hmmfetch %s %s >> %s\n' % (hmm, acc, hmm_fp)
        fa = '%s/fastas/%s.fa' % (odir, acc)
        fo = '%s/%s.fa' % (pfam_out, name)
        dia = '%s/%s.dmnd' % (pfam_out, name)
        if not isfile(dia):
            cmd += 'cp %s %s\n' % (fa, fo)
            cmd += 'diamond makedb --in %s -d %s\n' % (fo, dia)
            cmd += 'rm %s\n' % fo
        pfams[name] = [hmm_fp, dia]
    return pfams, cmd


def reads_lines(file_fp: str) -> list:
    """Check if file exists or is empty.

    Parameters
    ----------
    file_fp : str
        Input file.

    Returns
    -------
    file_lines : list
        Line of the input file.
    """
    file_lines = []
    if isfile(file_fp):
        with open(file_fp) as f:
            file_lines = [x.strip() for x in f.readlines()]
    return file_lines


def check_min_lines_count(input_fp: str) -> bool:
    """Check whether the number of lines in the file is above one.

    Parameters
    ----------
    input_fp : str
        Path to the input file.

    Returns
    -------
    ret : bool
        Whether the file contains at least one sequence.
    """
    ret = False
    with open(input_fp) as f:
        for ldx, line in enumerate(f):
            if ldx > 1:
                ret = True
                break
    return ret


def get_out_dir_(out_dir: str, inputs: dict, sam_pool: str,
                group: str = None) -> tuple:
    """Get the output directory for either the

    Parameters
    ----------
    out_dir : str
    inputs : dict
    sam_pool : str
        Name of the current sample or pool.
    group : str

    Returns
    -------
    out_dir : str
    file_path : str
    """
    inputs = inputs[sam_pool]
    if group:
        inputs = inputs[group]
    if out_dir.endswith('after_plass'):
        file_path = inputs.replace('nuclassembly', 'assembly')
    elif out_dir.endswith('after_prodigal'):
        file_path = inputs[1]
        if 'after_spades' in file_path:
            out_dir = out_dir.replace('after_', 'after_spades_')
        elif 'after_plass' in file_path:
            out_dir = out_dir.replace('after_', 'after_plass_')
    else:
        file_path = inputs[1]
    out_dir = '%s/%s' % (out_dir, sam_pool)
    return out_dir, file_path


def get_out_dir(self, sam_pool, group: str = None) -> tuple:
    """Get the output directory for either the

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder
        .inputs : dict
            Input files
    sam_pool : str
        Name of the current sample or pool.
    group : str

    Returns
    -------
    out_dir : str
    file_path : str
    """
    out_dir = self.dir
    inputs = self.inputs[sam_pool]
    if group:
        inputs = inputs[group]
    if self.dir.endswith('after_plass'):
        file_path = inputs.replace('nuclassembly', 'assembly')
    elif self.dir.endswith('after_prodigal'):
        file_path = inputs[1]
        if 'after_spades' in file_path:
            out_dir = out_dir.replace('after_', 'after_spades_')
        elif 'after_plass' in file_path:
            out_dir = out_dir.replace('after_', 'after_plass_')
    else:
        file_path = inputs[1]
    out_dir = '%s/%s' % (out_dir, sam_pool)
    return out_dir, file_path


def write_hmms(self) -> str:
    """Write a file to contain on each line, the paths to the .hmm
    files to search as part of integron_finder.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to the output directory.
        .databases.hmms_dias : dict
            .hmm files per name of profile.

    Returns
    -------
    hmms_fp : str
        Path to the file containing the paths to the .hmm files to search.
    """
    hmms_fp = ''
    if self.databases.hmms_dias:
        mkdr(self.dir)
        hmms_fp = '%s/hmm.txt' % self.dir
        o = open(hmms_fp, 'w')
        for target, gene_hmms in self.databases.hmms_dias.items():
            for gene, (hmm, _) in gene_hmms.items():
                o.write('%s\n' % hmm)
        o.close()
    return hmms_fp


def get_roundtrip(io) -> dict:
    roundtrip = {'to': inputs_to_scratch(io), 'from': outputs_back(io)}
    return roundtrip


def inputs_to_scratch(io) -> list:
    rsyncs, mkdirs = set(), set()
    # folders
    if ('I', 'd') in io:
        for folder_ in io[('I', 'd')]:
            folder = folder_.rstrip('/')
            src = folder_.rstrip('/').replace('${SCRATCH_FOLDER}', '')
            mkdirs.add('mkdir -p %s' % folder)
            rsyncs.add('rsync -aqruv %s/ %s' % (src, folder))
    # files
    if ('I', 'f') in io:
        for file in io[('I', 'f')]:
            folder = dirname(file)
            src = file.replace('${SCRATCH_FOLDER}', '')
            mkdirs.add('mkdir -p %s' % folder)
            rsyncs.add('rsync -aqruv %s %s' % (src, file))
    return sorted(mkdirs) + sorted(rsyncs)


def outputs_back(io) -> list:
    outbound = set()
    if ('O', 'd') in io:
        # folders
        for folder_ in io[('O', 'd')]:
            folder = folder_.rstrip('/')
            src = folder_.rstrip('/').replace('${SCRATCH_FOLDER}', '')
            cmd = 'mkdir -p %s; rsync -aqruv %s/ %s' % (src, folder, src)
            cmd = 'if [ -d %s ]; then %s' % (folder, cmd)
            outbound.add(cmd)
    if ('O', 'f') in io:
        # files
        for file in io[('O', 'f')]:
            src = file.replace('${SCRATCH_FOLDER}', '')
            folder = dirname(src)
            cmd = 'mkdir -p %s; rsync -aqruv %s %s' % (folder, file, src)
            cmd = 'if [ -f %s ]; then %s' % (file, cmd)
            outbound.add(cmd)
    return sorted(outbound)


def get_scratch_cmds(key, soft, cur_cmds, cmds):
    if key in soft.io:
        roundtrip = get_roundtrip(soft.io[key])
        scratched_cmds = ['# Move to SCRATCH_FOLDER'] + roundtrip['to']
        scratched_cmds += ['\n# %s commands (%s)' % (soft.name, key)] + cur_cmds
        scratched_cmds += ['\n# Move from SCRATCH_FOLDER'] + roundtrip['from']
        cmds[key] = scratched_cmds
    else:
        cmds[key] = cur_cmds


def per_coassembly_scratch(pool, soft, sam_cmds, cmds, commands):
    if pool in sam_cmds:
        get_scratch_cmds(pool, soft, sam_cmds, cmds)
    else:
        for group in commands.pools[pool]:
            group_cmds = sam_cmds[group]
            get_scratch_cmds((pool, group), soft, group_cmds, cmds)


def scratching(soft, commands) -> dict:
    if soft.params['scratch']:
        cmds = {}
        # Use commands.pools to unpack the commands well
        for sam, sam_cmds in soft.cmds.items():
            if isinstance(sam_cmds, list):
                get_scratch_cmds(sam, soft, sam_cmds, cmds)
            elif isinstance(sam_cmds, dict):
                per_coassembly_scratch(sam, soft, sam_cmds, cmds, commands)
            else:
                sys.exit('The collected commands are neither list of dict!')
        return cmds
    else:
        return soft.cmds


def io_update(self, i_f=None, i_d=None, o_f=None, o_d=None, key=None):

    for (IO_fd, val) in [
        (('I', 'f'), i_f),
        (('I', 'd'), i_d),
        (('O', 'f'), o_f),
        (('O', 'd'), o_d)
    ]:
        if not val:
            continue
        if isinstance(val, list):
            if key:
                self.outputs['io'][IO_fd].setdefault(key, set()).update(val)
            else:
                self.outputs['io'][IO_fd].update(val)
        elif isinstance(val, str):
            if key:
                self.outputs['io'][IO_fd].setdefault(key, set()).add(val)
            else:
                self.outputs['io'][IO_fd].add(val)




# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import re
import sys
import pandas as pd
from os.path import isfile, splitext


def run_import(input_path: str, output_path: str, typ: str) -> str:
    """
    Return the import qiime2 command.

    :param input_path: input file path.
    :param output_path: output file path.
    :param typ: qiime2 type.
    :return: command to qiime2.
    """
    cmd = ''
    if typ.startswith("FeatureTable"):
        if not input_path.endswith('biom'):
            cur_biom = '%s.biom' % splitext(input_path)[0]
            cmd += 'biom convert \\\n'
            cmd += '  -i %s \\\n' % input_path
            cmd += '  -o %s \\\n' % cur_biom
            cmd += '  --table-type="OTU table" \\\n'
            cmd += '  --to-hdf5\n\n'
            cmd += 'qiime tools import \\\n'
            cmd += '  --input-path %s \\\n' % cur_biom
            cmd += '  --output-path %s \\\n' % output_path
            cmd += '  --type "FeatureTable[Frequency]"\n'
        else:
            cmd += 'qiime tools import \\\n'
            cmd += '  --input-path %s \\\n' % input_path
            cmd += '  --output-path %s \\\n' % output_path
            cmd += '  --type "FeatureTable[Frequency]"\n'
    else:
        cmd += 'qiime tools import \\\n'
        cmd += '  --input-path %s \\\n' % input_path
        cmd += '  --output-path %s \\\n' % output_path
        cmd += '  --type "%s"\n' % typ
    return cmd


def run_export(input_path: str, output_path: str, typ: str) -> str:
    cmd = ''
    if typ.startswith("FeatureTable"):
        if not output_path.endswith('biom'):
            cur_biom = '%s.biom' % splitext(output_path)[0]
            cmd += 'qiime tools export \\\n'
            cmd += '  --input-path %s \\\n' % input_path
            cmd += '  --output-path %s\n' % splitext(output_path)[0]
            cmd += 'mv %s/*.biom %s\n' % (splitext(output_path)[0], cur_biom)
            cmd += 'biom convert'
            cmd += '  -i %s \\\n' % cur_biom
            cmd += '  -o %s.tmp \\\n' % output_path
            cmd += '  --to-tsv\n\n'
            cmd += 'tail -n +2 %s.tmp > %s\n\n' % (output_path, output_path)
            cmd += 'rm -rf %s %s.tmp\n' % (splitext(output_path)[0], output_path)
        else:
            cmd += 'qiime tools export \\\n'
            cmd += '  --input-path %s \\\n' % input_path
            cmd += '  --output-path %s\n' % splitext(output_path)[0]
            cmd += 'mv %s/*.biom %s\n' % (splitext(input_path)[0], output_path)
            cmd += 'rm -rf %s\n' % splitext(input_path)[0]
    else:
        cmd += 'qiime tools export \\\n'
        cmd += '  --input-path %s \\\n' % input_path
        cmd += '  --output-path %s\n' % splitext(output_path)[0]
        if 'songbird' in typ:
            cmd += 'mv %s/index.html %s\n' % (splitext(output_path)[0], output_path)
        else:
            cmd += 'mv %s/*.tsv %s\n' % (splitext(output_path)[0], output_path)
        cmd += 'rm -rf %s\n' % splitext(output_path)[0]
    return cmd


def write_seqs_fasta(out_fp_seqs_fasta: str, out_fp_seqs_qza: str,
                     tsv_pd: pd.DataFrame, tsv_fp: str = '') -> str:
    """
    Write the fasta sequences.

    :param out_fp_seqs_fasta: output sequences fasta file name.
    :param out_fp_seqs_qza: output sequences qiime2 Artefact file name.
    :param tsv_pd: table which feature names are sequences.
    :param cur_sh: writing file handle.
    """
    with open(out_fp_seqs_fasta, 'w') as fas_o:
        for seq in tsv_pd.index:
            fas_o.write('>%s\n%s\n' % (seq.strip(), seq.strip()))
    cmd = '# Write features as fasta file:\n'
    cmd += '#  - Features from: %s\n' % tsv_fp
    cmd += '# Snippet:\n'
    cmd += '# ```:\n'
    cmd += "# with open(fasta_out, 'w') as o:\n"
    cmd += "#     for seq in tsv_pd.index:\n"
    cmd += "#         o.write('>%s\\n%s\\n' % (seq.strip(), seq.strip()))\n"
    cmd += '# ```:\n'
    cmd += run_import(
        out_fp_seqs_fasta, out_fp_seqs_qza, 'FeatureData[Sequence]')
    return cmd


def write_taxonomy_sklearn(out_qza: str, out_fp_seqs_qza: str,
                           ref_classifier_qza: str) -> str:
    """
    Classify reads by taxon using a fitted classifier.
    https://docs.qiime2.org/2020.2/plugins/available/feature-classifier/classify-sklearn

    :param out_qza: Taxonomic classifications.
    :param out_fp_seqs_qza: The features sequences that should be classified.
    :param ref_classifier_qza: Taxonomic classifier.
    """
    cmd = 'qiime feature-classifier classify-sklearn \\\n'
    cmd += '--i-reads %s \\\n' % out_fp_seqs_qza
    cmd += '--i-classifier %s \\\n' % ref_classifier_qza
    cmd += '--p-n-jobs %s \\\n' % '4'
    cmd += '--o-classification %s\n' % out_qza
    return cmd


def get_case(
        case_vals: list,
        case_var: str,
        form: str = None) -> str:
    if len(case_vals):
        case = '%s_%s' % (case_var, '-'.join(
            [x.replace('<', 'below').replace('>', 'above') for x in case_vals]))
    else:
        case = case_var
    if form:
        case = '%s_%s' % (case, form)
    case = re.sub('[ /()]', '', case.replace('__', '_'))
    return case


def get_new_meta_pd(meta_pd: pd.DataFrame, case: str,
                    case_var: str, case_vals: list) -> pd.DataFrame:
    if 'ALL' in case:
        new_meta_pd = meta_pd.copy()
    elif len([x for x in case_vals if x[0] == '>' or x[0] == '<']):
        new_meta_pd = meta_pd.copy()
        for case_val in case_vals:
            if case_val[0] == '>':
                new_meta_pd = new_meta_pd[
                    new_meta_pd[case_var].astype(float) >= float(case_val[1:])
                ].copy()
            elif case_val[0] == '<':
                new_meta_pd = new_meta_pd[
                    new_meta_pd[case_var].astype(float) <= float(case_val[1:])
                ].copy()
    else:
        new_meta_pd = meta_pd[meta_pd[case_var].isin(case_vals)].copy()
    return new_meta_pd


def filter_feature_table(qza: str, new_qza: str, meta: str) -> str:
    cmd = '\nqiime feature-table filter-samples \\\n'
    cmd += '--i-table %s \\\n' % qza
    cmd += '--m-metadata-file %s \\\n' % meta
    cmd += '--o-filtered-table %s\n' % new_qza
    return cmd


def songbird_cmd(
        qza, new_qza, new_meta, nsams, params,
        formula, bformula, out_paths, force) -> tuple:

    fcmd = ''
    if force or not isfile(new_qza):
        fcmd += filter_feature_table(qza, new_qza, new_meta)

    batches = int(params['batches'])

    cmd = ''
    if force or not isfile(out_paths['diff_qza']):
        cmd = '# model\n'
        cmd += '\nqiime songbird multinomial \\\n'
        cmd += ' --i-table %s \\\n' % new_qza
        cmd += ' --m-metadata-file %s \\\n' % new_meta
        cmd += ' --p-formula "%s" \\\n' % formula
        cmd += ' --p-epochs %s \\\n' % params['epochs']
        if batches > 0.8 * nsams:
            cmd += ' --p-batch-size %s \\\n' % str(int(nsams * 0.8))
        else:
            cmd += ' --p-batch-size %s \\\n' % params['batches']
        cmd += ' --p-differential-prior %s \\\n' % params['diff_priors']
        cmd += ' --p-learning-rate %s \\\n' % params['learns']
        cmd += ' --p-min-sample-count %s \\\n' % params['thresh_samples']
        cmd += ' --p-min-feature-count %s \\\n' % params['thresh_feats']
        if 'examples' in params:
            cmd += ' --p-num-random-test-examples %s \\\n' % params['examples']
        else:
            cmd += ' --p-training-column %s \\\n' % params['train']
        cmd += ' --p-summary-interval %s \\\n' % params['summary_interval']
        cmd += ' --o-differentials %s \\\n' % out_paths['diff_qza']
        cmd += ' --o-regression-stats %s \\\n' % out_paths['stat']
        cmd += ' --o-regression-biplot %s\n' % out_paths['plot']

    if force or not isfile(out_paths['diff']):
        cmd += run_export(out_paths['diff_qza'], out_paths['diff'], '')

    stat_tsv = '%s.txt' % splitext(out_paths['stat'])[0]
    if force or not isfile(stat_tsv):
        cmd += run_export(out_paths['stat'], stat_tsv, '')

    bcmd = ''
    if force or len(out_paths['bdiff_qza']) and not isfile(out_paths['bstat']):
        bcmd += '\nqiime songbird multinomial \\\n'
        bcmd += ' --i-table %s \\\n' % new_qza
        bcmd += ' --m-metadata-file %s \\\n' % new_meta
        bcmd += ' --p-formula "%s" \\\n' % bformula
        bcmd += ' --p-epochs %s \\\n' % params['epochs']
        if batches > 0.8 * nsams:
            bcmd += ' --p-batch-size %s \\\n' % str(int(nsams * 0.8))
        else:
            bcmd += ' --p-batch-size %s \\\n' % params['batches']
        bcmd += ' --p-differential-prior %s \\\n' % params['diff_priors']
        bcmd += ' --p-learning-rate %s \\\n' % params['learns']
        bcmd += ' --p-min-sample-count %s \\\n' % params['thresh_samples']
        bcmd += ' --p-min-feature-count %s \\\n' % params['thresh_feats']
        if 'examples' in params:
            bcmd += ' --p-num-random-test-examples %s \\\n' % params['examples']
        else:
            bcmd += ' --p-training-column %s \\\n' % params['train']
        bcmd += ' --p-summary-interval %s \\\n' % params['summary_interval']
        bcmd += ' --o-differentials %s \\\n' % out_paths['bdiff_qza']
        bcmd += ' --o-regression-stats %s \\\n' % out_paths['bstat']
        bcmd += ' --o-regression-biplot %s\n' % out_paths['bplot']

    if force or not isfile(out_paths['tens']):
        cmd += '\nqiime songbird summarize-paired \\\n'
        cmd += ' --i-regression-stats %s \\\n' % out_paths['stat']
        cmd += ' --i-baseline-stats %s \\\n' % out_paths['bstat']
        cmd += ' --o-visualization %s\n' % out_paths['tens']

    if force or not isfile(out_paths['html']):
        cmd += run_export(out_paths['tens'], out_paths['html'], 'songbird')

    return cmd, fcmd, bcmd


def caller(self, namespace):
    """Calls as function the part of the software name that follows the first
    underscore.
    For example, software name "search_diamond" would call function`diamond()`.
    """
    func = self.soft.name.rsplit('_', 1)[1]
    module = sys.modules[namespace]
    if hasattr(module, func) and callable(getattr(module, func)):
        module_call = getattr(module, func)
        module_call(self)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from metagenomix.pipeline import Workflow
from metagenomix.config import AnalysesConfig
from metagenomix.databases import ReferenceDatabases
from metagenomix.commands import Commands
from metagenomix.jobs import CreateScripts


def metagenomix(**kwargs):
    """Run metagenomix to account for the passed commanbd line a.

    Parameters
    ----------
    kwargs : dict
        All arguments passed in command line, including defaults
    """

    # Parse the command line arguments and validate computing environment
    config = AnalysesConfig(**kwargs)
    config.run()

    # Validate and get commands to set up the passed databases if need be
    databases = ReferenceDatabases(config)
    databases.run()

    # Read and validate the workflow of tools to run as a pipeline
    workflow = Workflow(config, databases)
    workflow.run()

    # Collect the command line and  the workflow of tools to run as a pipeline
    commands = Commands(config, databases, workflow)
    commands.run()

    workflow.make_dirs()

    # Make .sh and scheduler (.slm or .pbs) scripts to
    scripting = CreateScripts(config)
    scripting.database_cmds(databases)  # build the databases
    scripting.software_cmds(commands)   # run the analysis pipeline
    scripting.display()  # show the scripts to run



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import subprocess
import numpy as np
from os.path import dirname, splitext

from metagenomix._io_utils import mkdr, scratching


class CreateScripts(object):

    def __init__(self, config):
        self.config = config
        self.cmd = []
        self.cmds = {}
        self.cmds_chunks = []
        self.soft = ''
        self.sh = ''
        self.run = {'database': {}, 'software': {}}
        self.job_fps = []
        self.job_name = ''
        self.modules = {}
        self.pjct = self.get_prjct()
        self.scheduler = self.get_scheduler()

    def get_scheduler(self):
        if self.config.jobs:
            if self.config.torque:
                return 'qsub'
            else:
                return 'sbatch'
        else:
            return 'sh'

    def get_prjct(self):
        prjct = [x for x in self.config.project if x.lower() not in 'aeiouy']
        if prjct:
            return ''.join(prjct)
        else:
            return self.config.project

    def get_cmds_chunks(self, n_chunks):
        if len(self.cmds) > int(n_chunks):
            cmds_split = np.array_split(list(self.cmds), n_chunks)
            self.cmds_chunks = [list(x) for x in cmds_split if len(x)]
        else:
            self.cmds_chunks = [[x] for x in self.cmds]

    def get_main_sh(self, name, soft=None) -> str:
        main = '%s/run_%s' % (self.sh.rsplit('/', 2)[0], name)
        if soft:
            main += '_after_%s.sh' % soft.prev
            self.run['software'][name] = main
        else:
            main += '.sh'
            self.run['database'][name] = main
        return main

    def write_main(self, name, soft=None) -> None:
        main = self.get_main_sh(name, soft)
        with open(main, 'w') as o:
            if len(self.job_fps):
                job_dir = dirname(self.job_fps[0])
                o.write('mkdir -p %s/output\n' % job_dir)
                o.write('cd %s/output\n' % job_dir)
                for job_fp in self.job_fps:
                    o.write('%s %s\n' % (self.scheduler, job_fp))
                self.job_fps = []

    def database_cmds(self, databases):
        for db, cmds in databases.commands.items():
            self.cmds = cmds
            self.get_cmds_chunks(self.config.params['chunks'])
            self.write_jobs(db)
            self.write_main(db)

    def get_modules(self, name: str):
        self.modules = self.config.modules.get(name, [])

    def software_cmds(self, commands):
        for sdx, (name, soft) in enumerate(commands.softs.items()):
            if not len(soft.cmds):
                continue
            print('[Writing commands] #%s: %s' % (sdx, soft.name))
            self.get_modules(name)
            self.cmds = scratching(soft, commands)
            self.get_cmds_chunks(soft.params['chunks'])
            self.write_jobs(name, soft)
            self.write_main(name, soft)

    def get_sh(self, name: str, cdx: int, soft=None) -> None:
        """

        Parameters
        ----------
        name : str
            Name of the current software of the pipeline workflow
        cdx
        soft

        Returns
        -------

        """
        if soft:
            self.sh = '%s/%s/jobs/run_%s_after_%s_%s.sh' % (
                self.config.dir, name, name, soft.prev, cdx)
        else:
            self.sh = '%s/databases/jobs/build_%s_%s.sh' % (
                self.config.dir, name, cdx)
        mkdr(dirname(self.sh))

    def prep_script(self, params: dict) -> None:
        # mandatory options
        self.cmd = [
            'Xhpc',
            '-j', self.job_name,
            '-t', str(params['time']),
            '-c', str(params['cpus']),
            '-M', str(params['mem_num']), params['mem_dim'],
            '--no-stat',
            '-i', self.sh]
        # whether the cpus requests is per node
        if params['nodes']:
            self.cmd.extend(['-n', str(params['nodes'])])
        # always provide an account
        if self.config.account:
            self.cmd.extend(['-a', self.config.account])
        # get the job script file path and use it
        job_script = '%s.slm' % splitext(self.sh)[0]
        if self.config.torque:
            self.cmd.append('--torque')
            job_script = '%s.pbs' % splitext(self.sh)[0]
        self.job_fps.append(job_script)
        self.cmd.extend(['-o', job_script])
        # whether an environment must be used for the current software
        if not self.modules and params['env']:
            self.cmd.extend(['-e', params['env']])
        # setup the scratch location to be used for the current software
        if isinstance(params['scratch'], int):
            self.cmd.extend(['--localscratch', str(params['scratch'])])
        elif params['scratch'] == 'scratch':
            self.cmd.append('--scratch')
        elif params['scratch'] == 'userscratch':
            self.cmd.append('--userscratch')
        # quiet Xhpc if metagenomix is not supposed to be verbose
        if not self.config.verbose:
            self.cmd.append('--quiet')
        # print(' '.join(self.cmd))

    def call_cmd(self):
        cmd = ' '.join(self.cmd)
        if self.config.verbose:
            print('[Running]', cmd)
        subprocess.call(cmd.split())
        # os.remove(self.sh)

    def write_chunks(self, chunks: list):
        with open(self.sh, 'w') as sh:
            for module in self.modules:
                sh.write('module load %s\n' % module)
            for chunk in chunks:
                for cmd in self.cmds[chunk]:
                    sh.write('%s\n' % cmd)

    def get_job_name(self, name: str, adx: int):
        self.job_name = name + '.' + self.pjct + '.' + str(adx)

    def write_script(self, soft=None):
        if self.config.jobs:
            if soft:
                self.prep_script(soft.params)
            else:
                self.prep_script(self.config.params)
            self.call_cmd()
        else:
            self.job_fps.append('%s.sh' % splitext(self.sh)[0])

    def write_jobs(self, name: str, soft=None):
        for cdx, chunks in enumerate(self.cmds_chunks):
            self.get_sh(name, cdx, soft)
            self.write_chunks(chunks)
            self.get_job_name(name, cdx)
            self.write_script(soft)

    def display(self):
        if len(self.run['database']) or len(self.run['software']):
            print()
            print('< PLEASE CONSIDER CHECKING THE SCRIPTS MANUALLY >')
        for database_software, name_main in self.run.items():
            print()
            print('# ========== #')
            print('  ', database_software)
            print('# ========== #')
            for name, main in name_main.items():
                print()
                print('>', name)
                print('sh', main)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from ._standalone_metagenomix import standalone_metagenomix

__all__ = ["metagenomix"]